from .api import hires_dino_256
from .api import hires_dino_512
from .api import hires_dino_1024
from .api import hires_dinov2

from .hires_dino import HighResDINO
from .lowres_dino import LowResDINO
